/**
 * Copyright (c) 2016 - 2019, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "nrf_dfu.h"

#include "nrf_dfu_utils.h"
#include "nrf_dfu_transport.h"
#include "nrf_dfu_req_handler.h"
#include "nrf_log.h"
#include "nrf_drv_qspi.h"
#include "nrf_dfu_serial.h"
#include "nrf_delay.h"
#include "nrf_dfu_flash.h"
#include "nrf_dfu_validation.h"
#include "nrf_dfu_settings.h"

//*********** Location of the images ***********
#define APPLICATION_INIT_PACKET_LOCATION 0x0000
#define APPLICATION_INIT_PACKET_MAX_SIZE 0x0200
#define APPLICATION_IMAGE_LOCATION 0x1000
#define APPLICATION_MAX_SIZE 0x20000
#define BOOTLOADER_SOFTDEVICE_INIT_PACKET_LOCATION 0x20000
#define BOOTLOADER_SOFDEVICE_IMAGE_LOCATION 0x21000
#define BOOTLOADER_SOFTDEVICE_MAX_SIZE 0x40000
//*******************************************
#define COMMAND_OBJECT              0x01 //For init packet 
#define DATA_OBJECT                 0x02 // For actual image packet
//*******************************************

uint16_t mtu_size = 200;
static nrf_dfu_observer_t m_user_observer;                          //<! Observer callback set by the user.
static uint32_t m_firmware_start_addr;          /**< Start address of the current firmware image. */
static uint32_t m_firmware_size_req;            /**< The size of the entire firmware image. Defined by the init command. */
/**
 * @brief This function calls the user's observer (@ref m_observer) after it is done handling the event.
 */
static void dfu_observer(nrf_dfu_evt_type_t event)
{
    switch (event)
    {
        case NRF_DFU_EVT_DFU_COMPLETED:
        {
#ifndef NRF_DFU_NO_TRANSPORT
            UNUSED_RETURN_VALUE(nrf_dfu_transports_close(NULL));
#endif
            break;
        }
        default:
            break;
    }

    /* Call user's observer if present. */
    if (m_user_observer)
    {
        m_user_observer(event);
    }
}

// Simple algorithm for finding the end of the firmware image in the external memory
uint32_t find_image_size(uint32_t start_address, size_t length)
{
    uint32_t img_size = 0;//start the image size at 0
    uint8_t m_rx_buf[length];
    nrf_drv_qspi_read(&m_rx_buf,length,start_address);//load init packet from flash into buffer
    nrf_delay_ms(50);
    while(img_size < length)
    {
        if(m_rx_buf[img_size] != 0xFF || m_rx_buf[img_size+1]!=0xFF || m_rx_buf[img_size+2]!=0xFF)
          //if the current byte and the next byte, are not cleared, go to the next byte
            img_size += 1;// keep count of the image size
        else
            break;//if the current and the next byte are cleared ( == 0xFF) then break and return image size
    }  
    
    return img_size;    
}

void init_pkt_transfer(void)
{
    nrf_dfu_response_t p_res;
    nrf_dfu_request_t p_req;

    uint32_t len = find_image_size(APPLICATION_INIT_PACKET_LOCATION,
                                               APPLICATION_INIT_PACKET_MAX_SIZE);
    uint8_t m_rx_buf[len];
    //Get MTU
    p_req.request = NRF_DFU_OP_MTU_GET;
    p_res.mtu.size = mtu_size;
    on_mtu_get_request(&p_req,&p_res);
    
    //select command
    p_req.request           = NRF_DFU_OP_OBJECT_SELECT;
    p_req.callback.response = dfu_req_handler_rsp_clbk;    
    p_req.select.object_type = COMMAND_OBJECT;
    nrf_dfu_command_req(&p_req,&p_res);
    
    //Create command
    p_req.request = NRF_DFU_OP_OBJECT_CREATE;
    p_req.create.object_type = COMMAND_OBJECT;
    p_req.create.object_size = len;
    nrf_dfu_command_req(&p_req,&p_res);

    //Write Command - writing from external to internal memory
    p_req.request = NRF_DFU_OP_OBJECT_WRITE;//change the opcode to write 
    nrf_drv_qspi_read(m_rx_buf,len+1,APPLICATION_INIT_PACKET_LOCATION);//read init packet from external memory into buffer starting at init packet location
    nrf_delay_ms(5);
    p_req.write.p_data = m_rx_buf;
    nrf_delay_ms(5);
    p_req.write.len = len;
    nrf_dfu_command_req(&p_req,&p_res);

    free((void*)p_req.write.p_data);

    //Calculate CRC Command
    p_req.request           = NRF_DFU_OP_CRC_GET; 
    nrf_dfu_command_req(&p_req,&p_res);

    //Execut Init packet Command
    p_req.request           = NRF_DFU_OP_OBJECT_EXECUTE; 
    nrf_dfu_command_req(&p_req,&p_res);

}

void fw_pkt_transfer(void)
{
    nrf_dfu_response_t p_res;
    nrf_dfu_request_t p_req;

    uint32_t len = find_image_size(APPLICATION_IMAGE_LOCATION, APPLICATION_IMAGE_LOCATION + APPLICATION_MAX_SIZE);

    uint8_t m_rx_buf[mtu_size];
    //select command
    p_req.request           = NRF_DFU_OP_OBJECT_SELECT;
    p_req.callback.response = dfu_req_handler_rsp_clbk;    
    p_req.select.object_type = DATA_OBJECT;
    nrf_dfu_data_req(&p_req,&p_res);
    
    //Create command
    p_req.request = NRF_DFU_OP_OBJECT_CREATE;
    p_req.create.object_type = DATA_OBJECT;
    p_req.create.object_size = len;
    nrf_dfu_data_req(&p_req,&p_res);

    //Write Command - writing from external to internal memory 
    uint32_t write_addr = APPLICATION_IMAGE_LOCATION;
    p_req.request = NRF_DFU_OP_OBJECT_WRITE;//change the opcode to write 
    while(s_dfu_settings.write_offset!=len)
    {
        memset(m_rx_buf,0,sizeof(m_rx_buf));//clear buffer
        if((s_dfu_settings.write_offset + mtu_size) < len)//if the current image size is less that the required image size
        {
            nrf_drv_qspi_read(m_rx_buf, mtu_size ,write_addr + s_dfu_settings.write_offset);//read init packet from external memory into buffer starting at init packet location
            nrf_delay_ms(5);//flush delay
            p_req.write.p_data = m_rx_buf;
            nrf_delay_ms(5);//flush delay
            p_req.write.len = mtu_size;
        }
        else
        {
            nrf_drv_qspi_read(m_rx_buf, len - s_dfu_settings.write_offset ,write_addr + s_dfu_settings.write_offset);//read init packet from external memory into buffer starting at init packet location
            nrf_delay_ms(5);
            p_req.write.p_data = m_rx_buf;
            nrf_delay_ms(5);
            p_req.write.len = len-s_dfu_settings.write_offset;
        }
        nrf_dfu_data_req(&p_req,&p_res);
        //printf("Next s_dfu_settings.write_offset size: %i\n",s_dfu_settings.write_offset);
        free((void*)p_req.write.p_data);
    }  
    //Calculate CRC Command
    p_req.request           = NRF_DFU_OP_CRC_GET; 
    nrf_dfu_data_req(&p_req,&p_res);

    //Execut Init packet Command
    p_req.request           = NRF_DFU_OP_OBJECT_EXECUTE; 
    nrf_dfu_data_req(&p_req,&p_res);
}

uint32_t nrf_dfu_init(nrf_dfu_observer_t observer)
{
    uint32_t ret_val;

    m_user_observer = observer;

    NRF_LOG_INFO("Entering DFU mode.");

    dfu_observer(NRF_DFU_EVT_DFU_INITIALIZED);
    ret_val = nrf_dfu_flash_init(false);

    if (ret_val != NRF_SUCCESS)
    {
        return ret_val;
    }

    nrf_dfu_validation_init();

    init_pkt_transfer();

    fw_pkt_transfer();

    return ret_val;
}
